
'''
Advanced Power Monitor
2008-2009(c) Kirill Plyashkevich <ru.spirit@gmail.com>
Connects to bunch of dbus signals, monitors values change and emmits signal in case of update.
'''

import random, gobject, os.path, signal, sys, time
import dbus, dbus.service
from dbus.mainloop.glib import DBusGMainLoop
from Util import *

class apm(dbus.service.Object):

  version = '0.5.1'
  iface = "ru.spirit.advanced_power_monitor"
  opath = "/ru/spirit/advanced_power_monitor"
  wrappers_root = 'wrappers'
  modules_root = 'modules'

  def __init__(self):
    self.m_loop = DBusGMainLoop()
    self.bus = dbus.SystemBus(mainloop = self.m_loop, private = True)
    self.bus_name = dbus.service.BusName(self.iface, self.bus) 
    dbus.service.Object.__init__(self, self.bus_name, self.opath)

    self.classname = '%s.%s' % (__name__, self.__class__.__name__)
    self.battery_is_charging= False
    self.battery_is_full = False
    self.battery_charger_is_disconnected = False
    self.wrappers = {}
    self.interfaces = {}
    self.signals = {}
    self.signal_queue = []
    self.min_queue = 1
    self.clients = set()
    self.clients_range = [0, 0xFFFFFFFF]
    self.working = False
    self.info_map = {}
    self.ifaces_map = {'CPUFreq':('org.freedesktop.Hal', '/org/freedesktop/Hal/devices/computer', 'org.freedesktop.Hal.Device.CPUFreq')}
    self.info_map = {'battery_status': {'value': 0, 'async': True, 'clients': set()},
                     'timeleft_idle': {'value': 0, 'async': True, 'clients': set()},
                     'timeleft_active': {'value': 0, 'async': True, 'clients': set()},
                     'device_mode': {'value': 'unknown', 'async': True, 'clients': set()},
                     'system_inactive': {'value': False, 'async': True, 'clients': set()},
                     'bluetooth_mode': {'value': 'unknown', 'async': True, 'clients': set()},
                     'bluetooth_modes': {'value': ['unknown'], 'async': True, 'clients': set()},
                     'usb_mode': {'value': 'unknown', 'async': True, 'clients': set()},
                     'usb_modes': {'value': ['unknown'], 'async': True, 'clients': set()}
                }

    print 'Loading modules...'
    modules = Module.get_modules_list(os.path.dirname(__file__), self.modules_root)
    del modules[self.modules_root]
    for k,v in modules.iteritems():
      Module.import_module(k)
      modules_list = Module.import_modules(v)
      class_list = []
      for m in modules_list:
        module_class = Module.get_module_class(m)
        if module_class:
          class_list.append(module_class)
      if class_list:
        for ref_module in class_list:
          c_module = ref_module()
          if c_module.available:
            print 'LOADED:\t%s' % (c_module.name)
            item = {'value': c_module.value, 'time': c_module.interval, 'object': c_module, 'async':c_module.async, 'clients': set()}
            self.info_map[k.replace(self.modules_root + '.', '')] = item
            break
          else:
            print 'FAILED:\t%s' % (c_module.name)
    print 'done\n'

    print 'Loading wrappers...'
    modules = Module.get_modules_list(os.path.dirname(__file__), self.wrappers_root)
    del modules[self.wrappers_root]
    for k,v in modules.iteritems():
      Module.import_module(k)
      modules_list = Module.import_modules(v)
      class_list = []
      for m in modules_list:
        module_class = Module.get_module_class(m)
        if module_class:
          class_list.append(module_class)
        else:
          del module_class
      if class_list:
        self.wrappers[k.replace(self.wrappers_root + '.', '')] = {'name': None, 'object': None, 'available': {}, 'list': class_list}
    self.init_wrappers()
    print 'done\n'

    self.init_interfaces()
    signal.signal(signal.SIGTERM, self._cleanup)
    self.working = True
    self.update_info()

  def _cleanup(self, *args):
    active_wrappers = self.wrappers.values()
    for wr in active_wrappers:
      try:
        wr['object']._cleanup()
      except KeyError:
        pass
    for x in self.info_map:
      try:
        gobject.source_remove(self.info_map[x]['src_evt'])
      except KeyError:
        pass
    self.destroy_interfaces()
    self.destroy_signals()
    self.working = False
    print 'EXIT'
    sys.exit(0)

#Signals subsystem
  def init_signals(self):
    for x in self.signals:
      self.add_signal(x)

  def add_signal(self, signal):
    self.remove_signal(signal)
    try:
      self.signals[signal]['obj'] = self.bus.add_signal_receiver(getattr(self, 'handle_%s' % signal), signal_name=self.signals[signal]['signal_name'], dbus_interface=self.signals[signal]['dbus_interface'], path=self.signals[signal]['path'])
    except Exception, e:
      self.print_exception('add_signal', str(e))

  def destroy_signals(self):
    for key in self.signals:
      self.remove_signal(key)

  def remove_signal(self, key):
    try:
      c_signal = self.signals[key]['obj']
      self.bus.remove_signal_receiver(c_signal)
      self.bus._clean_up_signal_match(c_signal)
      c_signal.remove()
      self.signals[key].pop('obj')
    except KeyError:
      pass
    except Exception, e:
      self.print_exception('remove_signal(%s)' % (key), str(e))
#end

#Interfaces subsystem
  def init_interfaces(self):
    for x in self.ifaces_map:
      try:
        self.interfaces[x] = dbus.Interface(self.bus.get_object(self.ifaces_map[x][0], self.ifaces_map[x][1]), self.ifaces_map[x][2])
      except Exception, e:
        self.interfaces[x] = None
        self.print_exception('init_interfaces', str(e))

  def destroy_interfaces(self):
    for key in self.interfaces:
      self.interfaces[key] = None
#end

#Wrappers routine
  def init_wrappers(self):
    for key in self.wrappers:
      self.check_wrapper(key)
      self.change_wrapper(key, self.get_available_wrappers(key)[0])
      print 'USING:\t%s\n' % str(self.wrappers[key]['name'])

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', out_signature='as')
  def get_wrappers_types(self):
    return self.wrappers.keys()

  def init_wrapper(self, key):
    signals = self.wrappers[key]['object'].get_signals()
    self.signals.update(signals)
    for s in signals:
      self.add_signal(s)

  def check_wrapper(self, key):
    if len(self.wrappers[key]['available']):
      self.wrappers[key]['available'].clear()
    ind = 0
    to_remove = set()
    for wr in self.wrappers[key]['list']:
      wrapper = wr(self.bus, True)
      name = wrapper.get_name()
      if wrapper.is_available():
        self.wrappers[key]['available'][name] = ind
        print 'LOADED:\t%s' % (name)
        ind += 1
      else:
        to_remove.add(wr)
        print 'FAILED:\t%s' % (name)
      wrapper._cleanup()
      del wrapper
    for wr in to_remove:
      self.wrappers[key]['list'].remove(wr)

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='ss')
  def change_wrapper(self, key, new_wrapper):
    wrapper_object = None
    try:
      wrapper_object = self.wrappers[key]['list'][self.wrappers[key]['available'][new_wrapper]](self.bus, True)
    except KeyError:
      return
    if new_wrapper != self.wrappers[key]['name'] and wrapper_object.is_available():#new_wrapper in self.wrappers[key]['available'].keys()
      if self.wrappers[key]['object']:
        old_wrapper_signals = self.wrappers[key]['object'].get_signals().keys()
        for sig_key in old_wrapper_signals:
          self.remove_signal(sig_key)
        self.wrappers[key]['object']._cleanup()
      self.wrappers[key]['name'] = new_wrapper
      wrapper_object.end_init()
      self.wrappers[key]['object'] = wrapper_object#self.wrappers[key]['list'][self.wrappers[key]['available'][new_wrapper]](self.bus)
      self.init_wrapper(key)
      if key == 'battery':
        self.update_percent()
        self.update_timeleft()
      elif key == 'bluetooth':
        try:
          self.set_value('bluetooth_mode', self.wrappers[key]['object'].get_mode())
          self.set_value('bluetooth_modes', self.wrappers[key]['object'].get_available_modes())
        except Exception, e:
          print e
      elif key == 'device_mode':
        try:
          self.set_value('device_mode', self.wrappers[key]['object'].get_device_mode())
          self.set_value('system_inactive', self.wrappers[key]['object'].get_system_inactive())
        except Exception, e:
          print e
      elif key == 'usb_mode':
        try:
          self.set_value('usb_mode', self.wrappers[key]['object'].get_mode())
          self.set_value('usb_modes', self.wrappers[key]['object'].get_available_modes())
        except Exception, e:
          print e

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='s', out_signature='as')
  def get_available_wrappers(self, key):
    result = self.wrappers[key]['available'].keys()
    if result:
      if result[0] == 'Dummy':
        result.append(result.pop(0))
    else:
      result.append(str(self.get_current_wrapper(key)))
    return result

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='s', out_signature='s')
  def get_current_wrapper(self, key):
    return str(self.wrappers[key]['name'])

#Client routine
  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', out_signature='x')
  def register(self):
    cl_min = self.clients_range[0]
    cl_max = self.clients_range[1]
    while True:
      client = random.randint(cl_min, cl_max)
      if not client in self.clients:
        self.clients.add(client)
        return client

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def unregister(self, client):
    for key in self.info_map:
      self.info_map[key]['clients'].discard(client)
    self.clients.discard(client)

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='xas')
  def subscribe(self, client, values=[]):
    client = long(client)
    if client in self.clients:
      for v in values:
        try:
          self.info_map[v]['clients'].add(client)
          c_key = self.info_map[v]
          if len(c_key['clients']) == 1 and 'object' in c_key:
              self.update_module(v)
          time.sleep(0.1)
        except KeyError:
          pass
      self.update_queue_size()

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='xas')
  def unsubscribe(self, client, values=[]):
    client = long(client)
    if client in self.clients:
      for v in values:
        try:
          self.info_map[v]['clients'].discard(client)
          if not self.info_map[v]['clients']:
            gobject.source_remove(self.info_map[v]['src_evt'])
        except KeyError:
          pass
      self.update_queue_size()

  def update_queue_size(self):
    info_map_values = self.info_map.values()
    info_map_async = [v for v in info_map_values if not v['async'] and len(v['clients'])]
    total = len(info_map_async)
    watching = sum([len(v['clients']) for v in info_map_async])
    self.min_queue = min(watching, 1 + watching/(total*len(self.clients)))

#Update information
  def update_info(self):
    for x in self.info_map:
      if not 'clients' in self.info_map[x]:
        self.info_map[x]['clients'] = set()
      if 'object' in self.info_map[x]:
        self.update_module(x)
    return self.working

  def update_module(self, key):
    try:
      self.check_property_modified([key, self.info_map[key]['object'].get_updated()], self.info_map[key]['object'].async)
    except Exception, e:
      self.print_exception('update_module(%s)' % (key), str(e))
    self.add_timer_event(key)
    return False

  def update_percent(self):
    self.check_property_modified(['battery_status', self.wrappers['battery']['object'].battery_percent], True)

  def update_timeleft(self):
    self.check_property_modified(['timeleft_idle', self.wrappers['battery']['object'].timeleft_idle], ['timeleft_active', self.wrappers['battery']['object'].timeleft_active], True)

#Battery signals handlers
  def handle_battery_percentage(self, *args):
    self.check_property_modified(['battery_status', self.wrappers['battery']['object'].get_percent(*args)], True)

  def handle_battery_timeleft_info(self, *args):
    timeleft_info = self.wrappers['battery']['object'].get_timeleft(*args)
    self.check_property_modified(['timeleft_idle', timeleft_info[0]], ['timeleft_active', timeleft_info[1]])

  def handle_battery_charger_is_connected(self):
    if not self.battery_is_charging:
      self.battery_is_charging= True
      self.battery_charger_is_disconnected = False
      self.PropertyModified(['battery_status'])

  def handle_battery_charger_is_disconnected(self):
    if not self.battery_charger_is_disconnected:
      self.battery_is_full = False
      self.battery_is_charging= False
      self.battery_charger_is_disconnected = True
      self.PropertyModified(['battery_status'])
      self.set_value('battery_status', -1)
    self.update_percent()

  def handle_battery_is_full(self):
    if not self.battery_is_full:
      self.battery_is_full = True
      self.PropertyModified(['battery_status'])

#Device mode signal handler
  def handle_device_mode(self, *args):
    self.wrappers['device_mode']['object'].device_mode_handler(*args)
    self.check_property_modified(['device_mode', self.wrappers['device_mode']['object'].get_device_mode()], True)

#System inactive signal handler
  def handle_system_inactive(self, *args):
    self.wrappers['device_mode']['object'].system_inactive_handler(*args)
    self.set_value('system_inactive', self.wrappers['device_mode']['object'].get_system_inactive())
    watching = [k for (k,v) in  self.info_map.iteritems() if len(v['clients']) and 'object' in v]
    if self.get_value('system_inactive'):
      for key in watching:
        try:
          gobject.source_remove(self.info_map[key]['src_evt'])
        except KeyError:
          pass
    else:
      for key in watching:
        self.update_module(key)

#Bluetooth mode signal handler
  def handle_bluetooth_mode(self, *args):
    self.wrappers['bluetooth']['object'].mode_handler(*args)
    self.check_property_modified(['bluetooth_mode', self.wrappers['bluetooth']['object'].get_mode()], True)

#USB mode signal handler
  def handle_usb_mode(self, *args):
    self.wrappers['usb_mode']['object'].mode_handler(*args)
    self.check_property_modified(['usb_mode', self.wrappers['usb_mode']['object'].get_mode()], True)

  @dbus.service.signal(dbus_interface='ru.spirit.advanced_power_monitor.signal', signature='as')
  def PropertyModified(self, keys):
    return keys

  def check_property_modified(self, *values):
    signals = []
    no_queue = False
    if values:
      values = list(values)
      if type(values[-1]) == bool:
        no_queue = values.pop(-1)
      for x in values:
        if self.get_value(x[0]) != x[1]:
          self.set_value(x[0], x[1])
          if len(self.info_map[x[0]]['clients']):
            signals.append(x[0])
      if signals:
        if no_queue:
          self.PropertyModified(signals)
        else:
          self.notify(signals)

  def notify(self, keys):
    watching = []
    for k in keys:
      try:
        clients_len = len(self.info_map[k]['clients'])
        if clients_len:
          self.signal_queue.append(k)
          watching.append(clients_len)
      except KeyError:
        pass
    info_map = self.info_map.values()
    total = len([v for v in info_map if not v['async'] and len(v['clients'])])
    c_queue = min(watching, 1 + sum(watching)/(total*len(self.clients)))
    if c_queue > self.min_queue:
      self.min_queue = c_queue
    result = set(self.signal_queue)
    if len(result) >= self.min_queue:
      self.signal_queue = []
      self.PropertyModified(result)

#Section: Battery state and percentage
#Key: battery_state
  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def get_status_info(self):
    try:
      self.wrappers['battery']['object'].get_status()
    except Exception, e:
      self.print_exception('get_status_info', str(e))

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', out_signature='s')
  def get_battery_state(self):
    result = 'unknown'
    if self.battery_is_full:
      result = 'full'
    elif self.battery_is_charging:
      result = 'charging'
    elif self.battery_charger_is_disconnected:
      result = 'charger_disconn'
      self.battery_charger_is_disconnected = False
    else:
      result = '%s%s' % (self.get_value('battery_status'), '%')
    return result
#end

#Section: Bluetooth
#Key: bluetooth_mode
  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='s')
  def set_bt_mode(self, mode):
    try:
      if self.wrappers['bluetooth']['object'] and mode != self.get_value('bluetooth_mode') and mode in self.get_value('bluetooth_modes'):
        if self.get_value('device_mode') == 'flight' and mode != 'off':
          self.set_device_mode('normal')
        self.wrappers['bluetooth']['object'].set_mode(mode)
    except Exception, e:
      self.print_exception('set_bt_mode', str(e))
#end

#Section: MCE
#Key: device_mode
  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def home_key_pressed_long(self):
    Shell.run_command_async('/usr/bin/dbus-send --system --type=signal /com/nokia/mce/signal com.nokia.mce.signal.sig_home_key_pressed_long_ind')

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def home_key_pressed(self):
    Shell.run_command_async('/usr/bin/dbus-send --system --type=signal /com/nokia/mce/signal com.nokia.mce.signal.sig_home_key_pressed_ind')

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def switch_device_mode(self):
    try:
      c_mode = self.get_value('device_mode')
      mode = c_mode
      if c_mode == 'normal':
        Shell.run_command_async('/usr/bin/dbus-send --system --dest=com.nokia.icd /com/nokia/icd_ui com.nokia.icd_ui.disconnect boolean:true')
        mode = 'flight'
      elif c_mode == 'flight':
        mode = 'normal'
      else:
        return
      self.set_device_mode(mode)
    except Exception, e:
      self.print_exception('switch_device_mode', str(e))

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='s')
  def set_device_mode(self, mode):
    try:
      if mode in ['flight', 'normal'] and self.wrappers['device_mode']['object']:
        self.wrappers['device_mode']['object'].set_device_mode(mode)
    except Exception, e:
      self.print_exception('set_device_mode', str(e))

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def tklock_device(self):
    try:
      if self.wrappers['device_mode']['object']:
        self.wrappers['device_mode']['object'].tklock()
    except Exception, e:
      self.print_exception ('tklock_device', str(e))

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def reboot(self):
    try:
      if self.wrappers['device_mode']['object']:
        self.wrappers['device_mode']['object'].reboot()
      else:
        Shell.run_command_async('/sbin/reboot -n')
    except Exception, e:
      self.print_exception('reboot', str(e))

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request')
  def shutdown(self):
    try:
      if self.wrappers['device_mode']['object']:
        self.wrappers['device_mode']['object'].shutdown()
      else:
        Shell.run_command_async('/sbin/shutdown -n')
    except Exception, e:
      self.print_exception('shutdown', str(e))
#end

#Section: CPU
#Key: cpu_governor, cpu_frequency
  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='s')
  def set_cpu_governor(self, governor):
    if not self.info_map['cpu_governor']['value'] == governor and governor in self.info_map['cpu_governors']['value']:
      try:
        self.interfaces['CPUFreq'].SetCPUFreqGovernor(governor)
        self.check_property_modified(['cpu_governor', self.interfaces['CPUFreq'].GetCPUFreqGovernor()])
      except KeyError:
        pass

#Section: USB Mode
#Key: usb_state
  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', in_signature='s')
  def set_usb_mode(self, mode):
    try:
      if self.wrappers['usb_mode']['object'] and mode != self.get_value('usb_mode') and mode in self.get_value('usb_modes'):
        self.wrappers['usb_mode']['object'].set_mode(mode)
    except Exception, e:
      self.print_exception('set_usb_mode', str(e))
#end

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', out_signature='s')
  def get_version(self):
    return self.version

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', out_signature='as')
  def get_property_list(self):
    return self.info_map.keys()

  @dbus.service.method(dbus_interface='ru.spirit.advanced_power_monitor.request', out_signature='v')
  def get_value(self, key):
    try: return self.info_map[str(key)]['value']
    except KeyError: return 'Error'

  def set_value(self, key, value=None):
    try: self.info_map[str(key)]['value'] = value
    except: pass

  def add_timer_event(self, key):
    if self.working and 'time' in self.info_map[key] and self.info_map[key]['time'] > 0 and len(self.info_map[key]['clients']) > 0 and not self.get_value('system_inactive'):
      self.info_map[key]['src_evt'] = gobject.timeout_add(self.info_map[key]['time'], self.update_module, key)

  def print_exception(self, func_name, message):
    print_exception_message('%s.%s()' % (self.classname, func_name), message)
