'''
Advanced Power Abstract Wrappers
2008-2009(c) Kirill Plyashkevich <ru.spirit@gmail.com>
'''

import dbus
import Util
from Util import System, Shell

class AbstractWrapper():

  def __init__(self, bus, check = False):
    self.bus = bus
    self.export_signals = {}
    self.export_interfaces = {}
    self.interfaces = {}
    self.interfaces_map = {}
    self.dep_packages = []
    self.friendly_name = ''
    self.available = False

    self.pre_init()
    self.check_available()

    if not check:
      self.end_init()

  def end_init(self):
    if self.available:
      self.init_interfaces()
      self.post_init()
      self.get_status()

  def pre_init(self):
    raise NotImplementedError("AbstractWrapper.pre_init() must be implemented")

  def post_init(self):
    raise NotImplementedError("AbstractWrapper.post_init() must be implemented")

  def get_status(self):
    raise NotImplementedError("AbstractWrapper.get_status() must be implemented")

  def _cleanup(self):
    self.destroy_interfaces()

  def get_dep_packages(self):
    return self.dep_packages

  def get_name(self):
    return self.friendly_name

  def get_signals(self):
    return self.export_signals

  def is_available(self):
    return self.available

  def check_available(self):
    self.available = System.are_packages_installed(self.dep_packages)
    return self.available

  def init_interfaces(self):
    for x in self.interfaces_map:
      try:
        self.interfaces[x] = dbus.Interface(self.bus.get_object(self.interfaces_map[x][0], self.interfaces_map[x][1]), self.interfaces_map[x][2])
      except:
        self.interfaces[x] = None

  def destroy_interfaces(self):
    for key in self.interfaces:
      self.interfaces[key] = None
